/*
 This Source Code Form is subject to the terms of the Mozilla Public
 License, v. 2.0. If a copy of the MPL was not distributed with this
 file, You can obtain one at http://mozilla.org/MPL/2.0/.

 GitHub Web Components Polyfill Add-on
 Copyright (c) 2020 JustOff. All rights reserved.
 Copyright (c) 2020 SeaHOH. All rights reserved.

 PerformanceObserver polyfill
 Copyright (c) 2021, Fastly, Inc. All rights reserved.
 https://github.com/fastly/performance-observer-polyfill

 Promise.any polyfill
 Copyright (c) 2019, Andrea Giammarchi, @WebReflection
 https://github.com/ungap/promise-any
 
 Custom Elements polyfill
 Copyright (c) 2020 The Polymer Project Authors. All rights reserved.
 https://github.com/webcomponents/polyfills
 
 Node.prototype.getRootNode polyfill
 Copyright (c) 2016 Foobar HQ
 https://github.com/foobarhq/get-root-node-polyfill
*/

"use strict";

var Cc = Components.classes, Ci = Components.interfaces, Cu = Components.utils;

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/services-crypto/utils.js");

Cu.importGlobalProperties(["XMLHttpRequest", "btoa"]);

//var console = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
//
//function log(msg) {
//  console.logStringMessage(msg);
//}

const polyfill = {
  getText: function (esc, url) {
    let xhr = new XMLHttpRequest();
    xhr.open("GET", url, false);
    xhr.responseType = "text";
    xhr.send(null);
    let text = xhr.responseText;
    if (esc)
        text = unescape(escape(text).replace(/%u/g, "\\u"));
    return text;
  },
  hashText: function (text) {
    let hash = btoa(CryptoUtils.UTF8AndSHA256(text));
    return `'sha256-${hash}'`;
  },
  urlsGithub: [
    [0, "chrome://github-wc-polyfill/content/performance-observer-polyfill.js"],
    // https://github.githubassets.com/assets/compat-838cedbb.js
    [1, "chrome://github-wc-polyfill/content/compat-838cedbb.js"],
  ],
  textGithub: `
(function () {
  window.queueMicrotask === undefined &&
      (queueMicrotask = (callback) => Promise.resolve().then(callback));
  Element.prototype.replaceChildren === undefined &&
      (Element.prototype.replaceChildren = function replaceChildren(...nodesOrDOMStrings) {
    while (this.lastChild) this.removeChild(this.lastChild);
    nodesOrDOMStrings !== undefined && this.append(...nodesOrDOMStrings);
  });
  Promise.any === undefined && (Promise.any = (function any($) {
    return new Promise(function (D, E, A, L) {
      A = [];
      L = $.map(function ($, i) {
        return Promise.resolve($).then(D, function (O) {
          return ((A[i] = O), --L) || E({errors: A});
        });
      }).length;
    });
  }).bind(Promise));
  String.prototype.replaceAll === undefined && (
      String.prototype.replaceAll = function replaceAll(pattern, replacement) {
    let isRe = pattern instanceof RegExp;
    if (!isRe)
      // $()*+,-.?[\\]^{|}
      pattern = pattern.replace(/[\\x24\\x28-\\x2e\\x3f\\x5b-\\x5e\\x7b-\\x7d]/g, "\\\\$\\&");
    if (!(isRe && pattern.global))
      pattern = new RegExp(pattern, "g");
    return this.replace(pattern, replacement);
  });
  // Ensure config dom.getRootNode.enabled is "false", or it would not work correctly
  if (Node.prototype.getRootNode === undefined) {
    Node.prototype.getRootNode = function getRootNode(opt) {
      let composed = typeof opt === "object" && Boolean(opt.composed);
      return composed ? getShadowIncludingRoot(this) : getRoot(this);
    }
    function getShadowIncludingRoot(node) {
      let root = getRoot(node);
      while (isShadowRoot(root))
        root = getRoot(root.host);
      return root;
    }
    function getRoot(node) {
      while (node.parentNode)
        node = node.parentNode;
      return node;
    }
    function isShadowRoot(node) {
      return node.nodeName === "#document-fragment" &&
             node.constructor.name === "ShadowRoot";
    }
  }
  // Remove "disabled" attribute of "Load diff" buttons
  if (/^\\/.+?\\/.+?\\/(commit\\/|pull\\/\\d+\\/files)/.test(location.pathname))
    document.addEventListener("DOMContentLoaded", function () {
      for (let button of document.getElementsByClassName("load-diff-button"))
        button.removeAttribute("disabled");
    }, {once: true});
}());
`,
  textFuGithub: `
(function () {
  // Idea from @misteuk https://github.com/JustOff/github-wc-polyfill/pull/48
  if (window.ShadowRoot === undefined) {
    ShadowRoot = class ShadowRoot extends DocumentFragment {
      set innerHTML (html) {
        super.innerHTML = html;
        let tagName = this.host.localName, hh = c32(tagName);
        // flag "s" is broken in matchAll
        for (let [css] of html.matchAll(/<style>[\\s\\S]+?<\\/style>/g)) {
          for (let [$, selectors, style] of css.matchAll(/(:host[\\s\\S]*?)(\\{[\\s\\S]+?\\})/g)) {
            let hc = c32($, hh);
            if (cssHashSet.has(hc)) continue
            cssHashSet.add(hc);
            CSS.innerHTML += selectors.replace(
                /:host(-context)?(?:\\((.+?)\\))?/sg,
                function ($, context, selectors) {
                  if (context === undefined && selectors === undefined)
                    return tagName;
                  if (context)
                    return \`:-moz-any(\${selectors}) \${tagName}\`;
                  let res = [];
                  for (let selector of selectors.split(","))
                    res.push(tagName + selector);
                  return \`:-moz-any(\${res.join(", ")})\`;
                }) + style;
          }
        }
      }
    };
    Element.prototype.attachShadow = function attachShadow(init) {
      if (this.shadowRoot !== undefined)
        throw new DOMException(
            \`The <\${this.tagName}> element has be tried to attach to is already a shadow host.\`,
            "InvalidStateError");
      if (!asNames.has(this.localName))
        throw new DOMException(
            \`The <\${this.tagName}> element does not supported to attach shadow\`,
            "NotSupportedError");
      let sr = new ShadowRoot();
      Object.defineProperty(sr, "host", {value: this});
      Object.defineProperty(sr, "mode", {value: init.mode});
      Object.defineProperty(sr, "delegatesFocus", {value: Boolean(init.delegatesFocus)});
      Object.defineProperty(this, "shadowRoot", {value: init.mode === "closed" ? null : sr});
      return sr;
    };
    function c32 (str, hash) {
      hash = hash || 0;
      for (let c of str) {
        if (" \\n\\t".includes(c)) continue
        hash += hash << 8;
        hash ^= c.codePointAt(0);
      }
      return hash;
    }
    function addEventListener (type, listener, options) {
      const target = this, parent = target.parentNode,
            ael = EventTarget.prototype.addEventListener;
      ael.call(target, type, listener, options);
      // Limit applied scope, or break other functions
      if (!(type === "click" && parent.localName === "button")) return;
      if (!triggerMap.has(target))
        triggerMap.set(target, function () {target.dispatchEvent(new Event("click"))});
      ael.call(parent, type, triggerMap.get(target), options);
    }
    function removeEventListener (type, listener, options) {
      const target = this, parent = target.parentNode,
            rel = EventTarget.prototype.removeEventListener;
      rel.call(target, type, listener, options);
      if (!(type === "click" && parent.localName === "button")) return;
      rel.call(parent, type, triggerMap.get(target), options);
    }
    const CSS = document.createElement("style"),
          cssHashSet = new Set(),
          triggerMap = new Map(),
          oldCED = customElements.define,
          asNames = new Set(["article", "aside", "blockquote", "body", "div",
                             "footer", "h1", "h2", "h3", "h4", "h5", "h6",
                             "header", "main", "nav", "p", "section", "span"]);
    document.head.appendChild(CSS);
    customElements.define = function (name, cls) {
      asNames.add(name);
      cls.prototype.addEventListener = addEventListener;
      cls.prototype.removeEventListener = removeEventListener;
      oldCED.call(customElements, name, cls);
    };
  }
}());
`,
  cssGithub: `<style></style>`,
  subGithubJs: [
    { // Nullish coalescing operator (??)
      "rePath": /\/vendors-node_modules_lit-html_lit-html_js-[0-9a-f]{20}\.js$/,
      "subList": [
        [/;(.)=(.)\?\?"";/, ';$1=$2!==undefined||$2!==null?$2:"";'],
      ]
    },
  ],
  urlsGitlab: [
    [0, "chrome://github-wc-polyfill/content/custom-elements.min.js"],
  ],
  textGitlab: "",
  subGitlabJs: [
    { // https://gitlab.com/gitlab-org/gitlab/-/commit/5522a5ffb6a0522e11dd684cce5c5e99ae6c24ce
      "rePath": /\/15\.[0-9a-f]{8}\.chunk\.js$/,
      "subList": [
        ["/^&(?<iid>\\d+)$/", "/^&(\\d+)$/"],
        [".groups.iid", "[1]"],
      ]
    },
    /*
    { // https://gitlab.com/gitlab-org/gitlab-ui/-/commit/d23c341bb8cef5683bf2d49556dabe4008be76cf
      // Test URL: https://gitlab.com/ykweyer-test-group/emoji-test
      "rePath":  /js$/,
      "subList": [
        ["\\p{Emoji}", "[\\u{1f300}-\\u{1f5ff}\\u{1f900}-\\u{1f9ff}\\u{1f600}-\\u{1f64f}\\u{1f680}-\\u{1f6ff}\\u{2600}-\\u{26ff}\\u{2700}-\\u{27bf}\\u{1f1e6}-\\u{1f1ff}\\u{1f191}-\\u{1f251}\\u{1f004}\\u{1f0cf}\\u{1f170}-\\u{1f171}\\u{1f17e}-\\u{1f17f}\\u{1f18e}\\u{3030}\\u{2b50}\\u{2b55}\\u{2934}-\\u{2935}\\u{2b05}-\\u{2b07}\\u{2b1b}-\\u{2b1c}\\u{3297}\\u{3299}\\u{303d}\\u{00a9}\\u{00ae}\\u{2122}\\u{23f3}\\u{24c2}\\u{23e9}-\\u{23ef}\\u{25b6}\\u{23f8}-\\u{23fa}]"],
      ]
    },
    */
    { // https://gitlab.com/gitlab-org/gitlab/-/merge_requests/79161
      "rePath":  /\/commons-pages\.admin\.topics\.edit-pages\.admin\.topics\.new-pages\.groups\.epics\.new-pages\.groups\.epics\.sho-[0-9a-f]{8}\.[0-9a-f]{8}\.chunk\.js$/,
      "subList": [
        ["/^(?<indent>\\s*)(?<leader>((?<isOl>[*+-])|(?<isUl>\\d+\\.))( \\[([x ])\\])?\\s)(?<content>.)?/",
         "/^(\\s*)(([*+-]|\\d+\\.)( \\[[x ]\\])?\\s)(.)?/"],
        [/\{indent:(.),content:(.),leader:(.)\}=(.)\.groups;/, "[$1,$2,$3]=[$4[1],$4[5],$4[2]];"],
      ]
    }
  ]
};

XPCOMUtils.defineLazyGetter(polyfill, "hashGithub", function () {
  for (let [esc, url] of this.urlsGithub)
    this.textGithub += this.getText(esc, url);
  this.textGithub += this.textFuGithub;
  this.subGithub = ["<link", `<script>${this.textGithub}</script>\n<link`];
  return this.hashText(this.textGithub);
});

XPCOMUtils.defineLazyGetter(polyfill, "hashGitlab", function () {
  for (let [esc, url] of this.urlsGitlab)
    this.textGitlab += this.getText(esc, url);
  this.subGitlab = ["<link", `<script>${this.textGitlab}</script>\n<link`];
  return this.hashText(this.textGitlab);
});

const domsGithub = new Set(`
github.com
gist.github.com
github.githubassets.com
`.split("\n"));
// Known GitLab instances based on https://gitlab.com/vednoc/dark-gitlab/-/blob/master/gitlab.user.css
const domsGitlab = new Set(`
gitlab.com
gitlab.net
0xacab.org
code.briarproject.org
code.foxkit.us
code.videolan.org
dev.gajim.org
devel.trisquel.info
forge.tedomum.net
foss.heptapod.net
framagit.org
git.alchemyviewer.org
git.callpipe.com
git.cardiff.ac.uk
git.cit.bcit.ca
git.coop
git.dev.ctu.univ-fcomte.fr
git.drk.sc
git.drupalcode.org
git.drupalcode.org
git.empiresmod.com
git.feneas.org
git.fosscommunity.in
git.gnu.io
git.happy-dev.fr
git.igwn.org
git.immc.ucl.ac.be
git.jami.net
git.ligo.org
git.linux-kernel.at
git.najer.info
git.nzoss.org.nz
git.oeru.org
git.ouru.org
git.pleroma.social
git.pwmt.org
git.regardscitoyens.org
git.rockylinux.org
git.silence.dev
git.synz.io
gitgud.io
gitplac.si
invent.kde.org
lab.libreho.st
mau.dev
mpeg.expert
opencode.net
repo.getmonero.org
salsa.debian.org
skylab.vc.h-brs.de
source.ind.ie
source.joinmastodon.org
source.puri.sm
source.small-tech.org
vbscan.fisica.unimib.it
`.split("\n"));
domsGithub.delete("");
domsGitlab.delete("");
const reGitlab = /^((next|www)\.)?gitlab\.(?!(io|biterg)[\.$])/;
const reGithubWs = /^\/assets-cdn\/worker\/socket-worker-[0-9a-f]+\.js$/;
const reGithubJs = /^https:\/\/github\.githubassets\.com\/assets\/.+\.js$/;
const reGitlabJs = /^\/assets\/webpack\/.+\.chunk\.js$/;
const reCspScript = /script-src [^;]+/;

const compat = {
  flags: 0,
  //reStop: /<title>.+?<\/title>/,
  _reStyle: '<link.+?integrity="(.+?)".+?name-([0-9a-f\\-]+)\\.css" />',
  _reScript: '<script.+?integrity="(.+?)".+?name-([0-9a-f\\-]+)\\.js"></script>',
  links: [],
  styles: [
    //{
    //  name: "frameworks",
    //  integrity: "sha512-VvfOEx8F6LRa1nh/1LEiw5jFLK0kJ+WE1LvuhfNDQkGC8vTGi5yYqCdPEPeBM9lNG7/pQ5cSU2/gOdE03JLCMg==",
    //  version: "56f7ce131f05e8b45ad6787fd4b122c3",
    //},
  ],
  javascripts: [
    {
      name: "vendors-node_modules_lit-html_lit-html_js",
      integrity: "",
      version: "$&",
    },
  ],
  _subRe: function (match, integrity, version) {
    return match.replace(integrity, this.integrity)
                .replace(version, this.version);
  },
  replace: function (text, done) {
    this.links.every(function (link) {
      if (done & link.flag) return true;
      let textRes = text.replace(link.reLink, link.subRe);
      if (textRes === text) return link.forward;
      text = textRes;
      done |= link.flag;
      return true;
    });
    //done !== this.flags && reStop.exec(text) && (done = this.flags)
    return [text, done];
  },
  init: function () {
    let reName, i = 0, filter = function (link) {
      this.flags |= (link.flag = 1 << i);
      link.reLink = new RegExp(this[reName].replace("name", link.name));
      link.subRe = this._subRe.bind(link);
      i = this.links.push(link);
    };
    reName = "_reStyle", this.styles.forEach(filter, this);
    reName = "_reScript", this.javascripts.forEach(filter, this);
    this.replace = this.replace.bind(this);
  }
};
compat.init();

var httpResponseObserver = {
  observe: function (subject, topic, data) {
    if (["http-on-examine-response", "http-on-examine-cached-response"].includes(topic) &&
        subject instanceof Ci.nsIHttpChannel && [200, 304].includes(subject.responseStatus)) {
      let URI = subject.URI,
          github = domsGithub.has(URI.host),
          gitlab = !github && (domsGitlab.has(URI.host) ||
                               Boolean(reGitlab.test(URI.host)));
      if (!(github || gitlab)) return;
      try {
        if (github && Boolean(reGithubWs.test(URI.path))) {
          let csp = subject.getResponseHeader("Content-Security-Policy");
          csp = csp.replace("worker-src ", "worker-src github.githubassets.com ");
          subject.setResponseHeader("Content-Security-Policy", csp, false);
        } else if (github && Boolean(reGithubJs.test(URI.spec)) ||
                   gitlab && Boolean(reGitlabJs.test(URI.path))) {
          subject.QueryInterface(Ci.nsITraceableChannel);
          let newListener = new tracingListener();
          newListener.site = github ? "githubjs" : "gitlabjs";
          newListener.path = URI.path;
          newListener.originalListener = subject.setNewListener(newListener);
        } else if (subject.responseStatus === 200 &&
            [Ci.nsIContentPolicy.TYPE_DOCUMENT, Ci.nsIContentPolicy.TYPE_SUBDOCUMENT]
            .includes(subject.loadInfo.externalContentPolicyType) &&
            subject.getResponseHeader("Content-Type").includes("text/html")) {
          if (URI.host !== "0xacab.org") {
            let csp = subject.getResponseHeader("Content-Security-Policy"),
                hash = github ? polyfill.hashGithub : polyfill.hashGitlab;
            csp = csp.replace(reCspScript, `$& ${hash}`);
            subject.setResponseHeader("Content-Security-Policy", csp, false);
          }
          subject.QueryInterface(Ci.nsITraceableChannel);
          let newListener = new tracingListener();
          newListener.site = github ? "github" : "gitlab";
          newListener.originalListener = subject.setNewListener(newListener);
        }
      } catch (e) {}
    }
  },
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver, Ci.nsISupportsWeakReference])
};

function textFromInputStream(inputStream, count) {
  let scriptableInputStream = Cc["@mozilla.org/scriptableinputstream;1"]
                              .createInstance(Ci.nsIScriptableInputStream);
  scriptableInputStream.init(inputStream);
  return scriptableInputStream.readBytes(count);
}

function inputStreamFromText(text) {
  let inputStream = Cc["@mozilla.org/io/string-input-stream;1"]
                    .createInstance(Ci.nsIStringInputStream);
  inputStream.setData(text, text.length);
  return inputStream;
}

function tracingListener() {}
tracingListener.prototype = {
  patched: 0,
  onDataAvailable: function (request, context, inputStream, offset, count) {
    if (!(this.patched & 1)) {
      let text;
      function replaceJs(subJs) {
        for (let {rePath, subList} of subJs) {
          if (!rePath.test(this.path)) continue
          for (let subs of subList) {
            if (text === undefined)
              text = textFromInputStream(inputStream, count);
            text = text.replace(...subs);
            }
        }
      }
      if (this.site === "github") {
        text = textFromInputStream(inputStream, count);
        // just a backup scheme
        let done = 0;
        if (!(this.patched & 1)) {
          [text, done] = compat.replace(text, done);
          done == compat.flags && (this.patched |= 1);
        }
        // ↓↓↓ set "this.patched |= 2" ↓↓↓ to use it
        if (!(this.patched & 2)) {
          text = text.replace(...polyfill.subGithub);
          this.patched |= 2;
        }
      }
      else if (this.site === "githubjs")
        replaceJs.call(this, polyfill.subGithubJs);
      else if (this.site === "gitlab") {
        text = textFromInputStream(inputStream, count);
        text = text.replace(...polyfill.subGitlab);
        this.patched = 1;
      }
      else if (this.site === "gitlabjs")
        replaceJs.call(this, polyfill.subGitlabJs);
      if (text !== undefined) {
        inputStream = inputStreamFromText(text);
        offset = 0;
        count = text.length;
      }
    }
    this.originalListener.onDataAvailable(request, context, inputStream, offset, count);
  },
  onStartRequest: function (request, context) {
    this.originalListener.onStartRequest(request, context);
  },
  onStopRequest: function (request, context, statusCode) {
    this.originalListener.onStopRequest(request, context, statusCode);
  },
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIStreamListener, Ci.nsISupports])
};

function startup(data, reason) {
  Services.obs.addObserver(httpResponseObserver, "http-on-examine-response", false);
  Services.obs.addObserver(httpResponseObserver, "http-on-examine-cached-response", false);
}

function shutdown(data, reason) {
  if (reason == APP_SHUTDOWN) {
    return;
  }
  Services.obs.removeObserver(httpResponseObserver, "http-on-examine-cached-response", false);
  Services.obs.removeObserver(httpResponseObserver, "http-on-examine-response", false);
}

function install() {}

function uninstall() {}
